#include <string.h>
#include "stdafx.h"
#include "match.h"
#include "PatMatch.h"

/* macros */

#define APPEL_OR        1
#define APPEL_AND       2
#define APPEL_NOT       4
#define APPEL_EXACT     8

/* types */

typedef struct CONNECTIVE_T {
    const char* name;
    int value;
} CONNECTIVE_T;

/* static data */
//static TREE_NODE_T *data_ref_el_tree = NULL;

/* functions */

/* convert a list of string to an APPEL_RULE_T */
int AppelMatcher::cvt_to_rule(GMap<GString, GString> &argv, Result &prule)
{

	if( GPosition pos=argv.contains("BEHAVIOR") )
	{
		if( argv[pos] == "request" )
			return A_BEHAVIOR_REQUEST;
	}
    /* convert */
    for (GPosition i = argv; i; ++i) 
	{
		const char* test=argv.key(i);
		const char* test2=argv[i];

        if ( argv.key(i) == "BEHAVIOR") 
		{
            /* found behavior */
            if (argv[i] == "request")
                prule.Flags |= P3PMATCH_REQUEST;
            else if (argv[i] == "limited")
                prule.Flags |= P3PMATCH_LIMITED;
            else if (argv[i] == "block")
                prule.Flags |= P3PMATCH_BLOCK;
            else
                return A_APPEL_PARSING_ERROR;
        } 
		else if (argv.key(i) == "CRTDBY") 
		{
            prule.CreatedBy += "\r\n" + argv[i];
        } 
		else if (strcmp(argv[i], "CRTDON") == 0) 
		{
            prule.CreatedOn += "\r\n" + argv[i + 1];
        } 
		else if (argv.key(i) == "DESCRIPTION") 
		{
            prule.MisMatches.append(argv[i]);
        } 
		else if (argv.key(i) == "PROMPT" 
                   && argv[i] == "yes") 
		{
            prule.Flags |= P3PMATCH_PROMPT;
        } 
		else if (argv.key(i) == "PERSONA") 
		{
            prule.Persona += "\r\n" +  argv[i];
        } 
		else if (argv.key(i) == "PROMPTMSG") 
		{
            prule.PromptMsgs.append(argv[i]);
        }
    }
    return A_OK;
}

/* match_attributes */
int AppelMatcher::match_attributes(GP<XmlNode> e, GP<XmlNode> x)
{

	for( GPosition epos=e->attrvalues; epos; ++epos)
	{
		if( e->attrvalues.key(epos) == "APPEL:CONNECTIVE" )
			continue;
		for( GPosition xpos=x->attrvalues; xpos && 
			e->attrvalues.key(epos) != x->attrvalues.key(xpos); ++xpos);
		/*if( !xpos || !epos || wild_match(x->attrvalues[xpos], e->attrvalues[epos]) == 0)
			return 0;*/
#ifdef _DEBUG
		if( xpos && epos )
		{
		const char* tmp1 = e->attrvalues[epos];
		const char* tmp2 = x->attrvalues[xpos];
		const char* tmp3 = e->attrvalues.key(epos);
		}
#endif
		if( !xpos || !epos || (!patternmatch(e->attrvalues[epos], x->attrvalues[xpos]) && !(e->name == "P3P:DATA" && e->attrvalues.key(epos) == "REF" && patternmatch(e->attrvalues[epos]+".*", x->attrvalues[xpos]))))
			return 0;
		
	}
	GString epcdata = e->content;
	GString xpcdata = x->content;
	epcdata.removewhite();
	xpcdata.removewhite();
	if( epcdata == xpcdata )
		return 1;
	else
		return 0;
}

/* pick off the connective */
int AppelMatcher::get_connective(GP<XmlNode> e)
{
    static CONNECTIVE_T connectives[] = {
        {"and", APPEL_AND},
        {"and-exact", APPEL_AND | APPEL_EXACT},
        {"non-and", APPEL_AND | APPEL_NOT},
        {"non-or", APPEL_OR | APPEL_NOT},
        {"or", APPEL_OR},
        {"or-exact", APPEL_OR | APPEL_EXACT},
        {NULL, 0},
    };

	for( GPosition epos=e->attrvalues; epos && 
		e->attrvalues.key(epos) != "APPEL:CONNECTIVE"; ++epos);
	
	if( !epos )
		return APPEL_AND;

	for (int j = 0;
         connectives[j].name != NULL
         && e->attrvalues[epos]!=connectives[j].name; j++);
    return connectives[j].value;
}

int AppelMatcher::match_request(GP<XmlNode> e, GURL &siteurl)
{
	if(e->name == "APPEL:REQUEST" )
	{
		if( GPosition uripos = e->attrvalues.contains("URI") )
		{
			if( patternmatch((const char*)e->attrvalues[uripos], (const char*)siteurl) )
				return 1;
		}
	}
	return 0;
}



/* appel_match_expression - match an expression in APPEL
 */
int AppelMatcher::appel_match_expression(GP<XmlNode> e, GP<XmlNode> x, GURL &siteurl)
{

#ifdef _DEBUG 
	const char *temp1 = e->name;
	const char *temp2 = x->name;

#endif

    int connective, didmatch;
    int matches, nonmatches;
    GP<XmlNode> echild, xchild;

    if (e->name == "APPEL:OTHERWISE")
        return 1;

	if (e->name != "APPEL:RULE" && e->name != "APPEL:REQUEST-GROUP")
	{
		/*
		   The standard matching semantics for rules in APPEL are as follows:

		   An expression E matches a piece of evidence X (i.e. a certain XML element
		   in the evidence) if and only if all of the following holds:

		   1. The element names of E and X are identical
		 */
		if (e->name != x->name
			&& e->name.substr(0, 4) != "P3P:"
				|| e->name.substr(4, -1) != x->name )
			return 0;
		/*
		   2. All of E's attribute expressions match attributes of X (additional
		   attributes in evidence X which are not referenced in expression E
		   are ignored)
		 */
		if (match_attributes(e, x) == 0)
			return 0;
	}

    /* gkp: Here is a good place for optimization of comparisons */
    connective = get_connective(e);

	GPosition epos = e->first_child();
	matches = nonmatches = 0;
	while( epos )
	{
		echild = e->next_child(epos);
		didmatch = 0;

		if (e->name == "APPEL:RULE" )
		{
			if( appel_match_expression(echild, x, siteurl) )
					didmatch = 1;
		}
		else if( e->name == "APPEL:REQUEST-GROUP" )
		{
			if( match_request(echild, siteurl) )
				didmatch = 1;
		}
		else
		{
			GPosition xpos = x->first_child();
			while( xpos   )
			{
				xchild = x->next_child(xpos);
				if( appel_match_expression(echild, xchild, siteurl) )
				{
					didmatch = 1;
					break;
				}
			}
		}

		if( didmatch )
			matches++;
		else
			nonmatches++;
	}

	/////////////////////////////////////////////////////////////

    /*
       3. [if an or connective is given in E] at least one of E's contained
       expressions (if any) match X's enclosed elements (additional
       enclosed elements in evidence X which are not referenced in
       expression E are ignored).
     */
    if (connective == APPEL_OR)
        return (matches > 0);
    /*
       4. [if an and connective is given in E] all of E's contained expressions
       (if any) match X's enclosed elements (additional enclosed elements
       in evidence X which are not references in expression E are ignored).
     */
    else if (connective == APPEL_AND)
        return (nonmatches == 0);
    /*
       5. [if an non-or connective is given in E] none of E's contained 
       expressions (if any) match X's enclosed elements (additional enclosed 
       elements in evidence X which are not references in expression E are 
       ignored).
     */
    else if (connective == (APPEL_OR | APPEL_NOT))
        return (matches == 0);
    /*
       6. [if an non-and connective is given in E] not all of E's contained 
       expressions (if any) match X's enclosed elements (additional enclosed 
       elements in evidence X which are not references in expression E are 
       ignored).
     */
    else if (connective == (APPEL_AND | APPEL_NOT))
        return (nonmatches > 0);
    /*
       7. [if an or-exact connective is given in E] at least one of E's
       contained expressions (if any) match X's enclosed elements (additional
       enclosed elements in evidence X which are not referenced in expression E
       are not ignored).
     */
    else if (connective == (APPEL_OR | APPEL_EXACT))
        return (matches == x->num_children());
    /*
       8. [if an and-exact connective, or no connective is given in E] all of 
       E's contained expressions (if any) match X's enclosed elements 
       (additional enclosed elements in evidence X which are not referenced in 
       expression E are not ignored).
     */
    else
        return (matches == e->num_children()
                && matches == x->num_children());
    return 1;
}

/* appel_match -
   returns A_OK if we found a match, an error otherwise.
 */
int AppelMatcher::appel_match(GP<XmlNode> appel_tree, GP<XmlNode> p3p_tree,
                Result &appel_rule)
{
	appel_rule.Match = Result::UNKNOWN;

    GP<XmlNode> xmldata;
    GP<XmlNode> treenode;

    //if (data_ref_el_tree == NULL)
      //  read_xml_data(); 

    if (appel_tree->num_children() == 0) {
        /* a truly degenerative case -- no nodes at top of tree */
        return A_OK;
    }

    /* assert that it is the "appel:RULESET" node) */
    xmldata = appel_tree;
    if ( xmldata->name != "APPEL:RULESET" ) 
	{
		appel_rule.Match = Result::UNKNOWN;
        return A_APPEL_PARSING_ERROR;
    }
    /* are there any rules? */
    if (appel_tree->num_children() == 0) {
        return A_OK;
    }
    /* fix the P3P tree */
    if (p3p_tree->num_children() == 0) {
        /* no nodes in P3P */
		appel_rule.Match = Result::UNKNOWN;
        return A_P3P_PARSING_ERROR;
    }

    xmldata = p3p_tree;
    if (xmldata->name != "POLICY" ) {
		appel_rule.Match = Result::UNKNOWN;
        return A_P3P_PARSING_ERROR;
    }
	else
	{
		
		for(GPosition i=xmldata->attrvalues; i ; ++i )
		{
			if( xmldata->attrvalues.key(i) == "DISCURI"  )
			{
				appel_rule.DiscURI = xmldata->attrvalues[i];
			}
			else if( xmldata->attrvalues.key(i) == "OPTURI" )
			{
				appel_rule.OptURI = xmldata->attrvalues[i];
			}

		}
	}


	GPosition tpos = appel_tree->first_child();
	treenode = appel_tree->next_child(tpos);
	while(treenode)
	{
#ifdef _DEBUG
		const char *nodename = (const char*)treenode->name;
#endif
		//while(treenode && appel_match_expression(treenode->next_child(treenode->first_child()), p3p_tree) == 0)
		while(treenode && appel_match_expression(treenode, p3p_tree, appel_rule.resultforURL) == 0)
		{
			treenode = appel_tree->next_child(tpos);
		}

		if( !treenode )
			break;

		if( cvt_to_rule( treenode->attrvalues, appel_rule) == A_BEHAVIOR_REQUEST)
			break;
		treenode = appel_tree->next_child(tpos);
	}
	if( appel_rule.MisMatches.size() )
		appel_rule.Match = Result::UNMATCH;
	else
		appel_rule.Match = Result::MATCH;

	return A_OK;
}


/* find a policy tag in a tree given a name - 
   return NULL if not found
 */
GP<XmlNode> AppelMatcher::find_policy_tag(GP<XmlNode> root, GString name)
{
    GP<XmlNode> child, node;
    GP<XmlNode> xmldata;

    if (root == NULL || name == NULL)
        return NULL;
    xmldata = root;
    if ( xmldata->name == "POLICY") 
	{
        /* found a policy tag; is it the right one? */
        for (GPosition i = xmldata->attrvalues;
             i && xmldata->attrvalues.key(i) != "NAME"; ++i);

        if (!i)
            return NULL;
        if (xmldata->attrvalues[i] == name)
            return root;
        return NULL;
    }
    /* wasn't this node? check its children */
	GPosition pos = root->first_child();
	while( pos )
	{
		child = root->next_child(pos);
		node = find_policy_tag(child, name);
		if( node )
			return node;
	}

    return NULL;
}

/* add defaults to P3P tree */
void AppelMatcher::add_p3p_defaults(GP<XmlNode> p3p_tree)
{
    GP<XmlNode> child;
    GP<XmlNode> xmldata;

    if (p3p_tree == NULL)
        return;
    xmldata = p3p_tree;
    if (xmldata != NULL)
        xml_add_p3p_defaults(xmldata);
    /* wasn't this node? check its children */
	GPosition pos = p3p_tree->first_child();
	while(pos)
	{
		child = p3p_tree->next_child(pos);
		add_p3p_defaults(child);
	}

}

static int p3p_compare(const void *a, const void *b)
{
    char *pa = (char *) a;
    char *pb = *(char **) b;

    return strcmp(pa, pb);
}


void AppelMatcher::xml_add_p3p_defaults(GP<XmlNode> xmldata)
{
    static const char *p3p_defaults[] = {
        "ADMIN", "REQUIRED", "always",
        "CONTACT", "REQUIRED", "always",
        "CURRENT", "REQUIRED", "always",
        "CUSTOMIZATION", "REQUIRED", "always",
        "DATA", "OPTIONAL", "no",
        "DATA-GROUP", "BASE", "http://www.w3.org/TR/P3P/base",
        "DELIVERY", "REQUIRED", "always",
        "DEVELOP", "REQUIRED", "always",
        "EXTENSION", "OPTIONAL", "yes",
        "HISTORICAL", "REQUIRED", "always",
        "INDIVIDUAL-ANALYSIS", "REQUIRED", "always",
        "INDIVIDUAL-DECISION", "REQUIRED", "always",
        "OTHER-PURPOSE", "REQUIRED", "always",
        "OTHER-RECIPIENT", "REQUIRED", "always",
        "PSEUDO-ANALYSIS", "REQUIRED", "always",
        "PSEUDO-DECISION", "REQUIRED", "always",
        "PUBLIC", "REQUIRED", "always",
        "SAME", "REQUIRED", "always",
        "TAILORING", "REQUIRED", "always",
        "TELEMARKETING", "REQUIRED", "always",
        "UNRELATED", "REQUIRED", "always",
    };
    char **s;

    /* find name in list; if not there, return */
    s = (char**)bsearch(xmldata->name, p3p_defaults,
                sizeof(p3p_defaults) / (3 * sizeof(char *)),
                3 * sizeof(char *), p3p_compare);
    if (s == NULL)
        return;
    /* if the attribute is explicit, return */
    for (GPosition i = xmldata->attrvalues; i; ++i)
        if ( xmldata->attrvalues.key(i) == s[1] )
            return;
    /* add the default attribute */
	xmldata->attrvalues[s[1]] = s[2];

}

int AppelMatcher::match(GP<XmlNode> appel_tree, GP<XmlNode> p3p_tree,
                Result &result)
{
	add_p3p_defaults(p3p_tree);
    /* match p3p_tree against appel_tree */
    return appel_match(appel_tree, p3p_tree, result);
}


/* END OF FILE */
