//C- Copyright (c) 2001 AT&T Corp.
//C- All rights reserved.
//C- $Id: PolicyHandler.cpp,v 1.40 2003/02/12 22:42:23 praveen Exp $

#include "stdafx.h"
#include "PolicyHandler.h"
#include "globals.h"
#include "PatMatch.h"
#include "ReadHTTPFile.h"
#include "P3PPreferences.h"
#include "XmlFile.h"
#include "GOS.h"
#include <direct.h>
#include "SaxXmlParse.h"

PolicyReferenceHandler g_policy_handler;

GMap<GURL, GP<SitePolicies> > PolicyReferenceHandler::site_policy_reference_map;
GMap<GURL, GP<CacheData> > PolicyReferenceHandler::policy_cache;
GMap<GString, GString> PolicyReferenceHandler::hint_map;
GMap<GURL, bool> PolicyReferenceHandler::downloaded_urls;

GCriticalSection PolicyReferenceHandler::class_lock;


void PolicyReferenceHandler::ClearCache()
{
	GCriticalSectionLock lock(&class_lock);
	policy_cache.empty();
	GOS::cleardir(g_dllpath+"\\cache");
	site_policy_reference_map.empty();
	hint_map.empty();
}


void PolicyReferenceHandler::SavePolicyCache()
{
	LoadPolicyCache();  /// to merge any other entries

	GCriticalSectionLock lock(&class_lock);
	GString cachedir = g_dllpath+"\\cache";
	_mkdir(cachedir);

	XmlOutFile xmlf(cachedir+"\\cacheindex.xml");
	XmlOutNode root(xmlf);
	root.setNodeName("POLICY-CACHE");

	for(GPosition pos=policy_cache; pos; ++pos)
	{
		GP<CacheData> cd = policy_cache[pos];
		
		XmlOutNode node(xmlf, true);
		GMap<GString, GString> attr;
		attr["URL"] = (const char*)policy_cache.key(pos);
		attr["EXPIRY"] = GString(policy_cache[pos]->expiry);
		attr["FILENAME"] = policy_cache[pos]->filename;
		node.setNodeName("CACHE-ENTRY", attr);
	}
}


void PolicyReferenceHandler::LoadPolicyCache()
{
	GCriticalSectionLock lock(&class_lock);

	time_t t = time(NULL);
	GString cachedir = g_dllpath+"\\cache\\";
	TRY
	{
		GP<XmlNode> root = SaxXmlParse(cachedir+"cacheindex.xml", false);//XMLParse(cachedir+"cacheindex.xml");
		if( root && root->name == "POLICY-CACHE")
		{
			GPosition pos;
			GP<XmlNode> cacheentry = root->get_first_child(pos);
			while(cacheentry)
			{
				if( cacheentry->name == "CACHE-ENTRY" )
				{
					
					if( GPosition tmppos = cacheentry->attrvalues.contains("URL") )
					{
						TRY
						{
							GURL url;
							url = cacheentry->attrvalues[tmppos];
							if( tmppos = cacheentry->attrvalues.contains("FILENAME") )
							{
								GString filename = cacheentry->attrvalues[tmppos];
								if( !filename.length() || GOS::is_file(cachedir+filename) )
								{
									if( tmppos = cacheentry->attrvalues.contains("EXPIRY") )
									{
										time_t expiry = atol(cacheentry->attrvalues[tmppos]);
										if( expiry > t )
										{
											if( GPosition polpos = policy_cache.contains(url) )
											{
												if( policy_cache[polpos]->filename != filename )
													unlink(cachedir+filename);
											}
											else
											{
												GP<CacheData> cachedata = new CacheData(filename, expiry);
												policy_cache[url] = cachedata;
											}
										}
										else
											unlink(cachedir+filename);
									}
									else
										unlink(cachedir+filename);
								}
							}
						}
						CATCH(ex)
						{
						}
						ENDCATCH;

					}
				}
				cacheentry = root->get_next_child(pos);
			}  ///while
		}  /// if
	}
	CATCH(ex)
	{
	}
	ENDCATCH;
}

time_t GetExpiryFromPolicyData(const char* data)
{
	if(!data)
		return 0;
	TRY
	{
		GP<XmlNode> root = SaxXmlParse(data, strlen(data), false);//XMLParse(data, strlen(data) );
		if( root )
		{
			if( root->name == "META" )
			{	GPosition pos;
				root = root->get_first_child(pos);
				if( !root )
					return 0;
			}
			const char* data = root->name;
			if( root->name == "POLICY-REFERENCES" || root->name == "POLICIES" || root->name == "POLICY")
			{
				GPosition pos;
				GP<XmlNode> node = root->get_first_child(pos);
				while(node)
				{
					data = node->name;
					if( node->name == "EXPIRY" )
					{
						GPosition pos;
						if(pos = node->attrvalues.contains("DATE"))
						{
							GString date = node->attrvalues[pos];
							return InternetTimeToUnixTime(date);
						}
						else if( pos = node->attrvalues.contains("MAX-AGE") )
						{
							GString age = node->attrvalues[pos];
							return time(NULL)+atol(age);
						}
						return 0;
					}
					node = root->get_next_child(pos);
				}
			}
		}
	}
	CATCH(ex)
	{
		return -1;
	}
	ENDCATCH;
	return 0;
}



		

bool PolicyReferenceHandler::CacheURL(const GURL &policyurl, const char *policy, time_t expiry)
{
	GCriticalSectionLock lock(&class_lock);
	AddDownloadedURL(policyurl);

	if( !policy )
	{ // dummy cache for one day
		if( !expiry )
			expiry = time(NULL)+60*60*24;

		GP<CacheData> data = new CacheData("", expiry);
		policy_cache[policyurl] = data;
		return true;
	}

	if( !expiry )
		expiry = GetExpiryFromPolicyData(policy);
	if( !expiry )
		expiry = time(NULL)+60*60*24;
	else if( expiry < time(NULL) )
		return false;

	GString cachedir = g_dllpath+"\\cache";
	_mkdir(cachedir);
	GString filename = tempfilename(cachedir);
	FILE *fp;
	if( fp = fopen(cachedir+"\\"+filename, "wb") )
	{
		fwrite((const char*)policy, strlen(policy), 1, fp);
		fclose(fp);
	}

	GP<CacheData> data = new CacheData(filename, expiry);

	policy_cache[policyurl] = data;
	return true;
}

GString PolicyReferenceHandler::GetLocalPathFromCache(const GURL &policyurl)
{
	GCriticalSectionLock lock(&class_lock);

	GURL url = policyurl;
	SeparateNameAndURL(url);

	if( GPosition pos = policy_cache.contains(url) )
	{
		return g_dllpath+"\\cache\\"+policy_cache[pos]->filename;
	}
	return GString();
}

char *PolicyReferenceHandler::GetURLFromCache(const GURL &policyurl, bool deleteifexpired)
{
	GCriticalSectionLock lock(&class_lock);

	if( GPosition pos = policy_cache.contains(policyurl) )
	{
		if( policy_cache[pos]->expiry < time(NULL) && deleteifexpired )
		{
			unlink(g_dllpath+"\\cache\\"+policy_cache[pos]->filename);
			policy_cache.del(pos);
			return NULL;
		}
		else
		{

			if( !policy_cache[pos]->filename.length() )
				return NULL; /// dummy cache entry to reflect download

			FILE *fp;
			if( fp = fopen(g_dllpath+"\\cache\\"+policy_cache[pos]->filename, "rb") )
			{
				fseek(fp, 0, SEEK_END);
				int length = ftell(fp);
				fseek(fp, 0, SEEK_SET);
				char *buf = new char[length+1];
				fread(buf, length, 1, fp);
				buf[length] = 0;
				fclose(fp);
				return buf;
			}
			else
			{
				policy_cache.del(pos);
				return NULL;
			}
		}
	}
	else
		return NULL; // return null
}


bool PolicyReferenceHandler::isURLExistsInCache(const GURL &policyurl)
{
	GCriticalSectionLock lock(&class_lock);

	if( GPosition pos = policy_cache.contains(policyurl) )
	{
		if( policy_cache[pos]->expiry < time(NULL) )
		{
			unlink(g_dllpath+"\\cache\\"+policy_cache[pos]->filename);
			policy_cache.del(pos);
			return false;
		}
		else if( !policy_cache[pos]->filename.length() || GOS::is_file(g_dllpath+"\\cache\\"+policy_cache[pos]->filename) )
			return true;
		else
		{
			policy_cache.del(pos);
			return false;
		}

	}
	else
		return false;
}

char *PolicyReferenceHandler::GetURL(const GURL &realurl, GString &header, bool download, bool deleteifexpired)
{
	if( realurl.is_empty() )
		return NULL;

	GURL url = realurl;
	SeparateNameAndURL(url);

	if( isURLExistsInCache(url) )
	{
		return GetURLFromCache(url, deleteifexpired);
	}
	else if( download )
	{
		char *data = GetPage(url, header);
		if( CacheURL(url, data) )
			return data;
	}

	return NULL;
}

char *PolicyReferenceHandler::GetURL(const GURL &url, bool download, bool deleteifexpired)
{
	GString header;
	return GetURL(url, header, download, deleteifexpired);
}

GURL PolicyReferenceHandler::FindPolicyForCookie(const GString &name, const GString &value, GString domain, GString path)
{
	GCriticalSectionLock lock(&class_lock);
	
	if( !domain.length() )
		return GURL();

	GString stardomain = "*."+domain;
	for( GPosition pos = site_policy_reference_map; pos; ++pos )
	{
		if( patternmatch(stardomain, site_policy_reference_map.key(pos).domain() ) )
		{
			GP<SitePolicies> sp = site_policy_reference_map[pos];
			for( GPosition refpos=sp->sitepolicyrefs; refpos; ++refpos)
			{///go through each policy reference to find out we have the required policy url

				GP<PolicyReference> pr  = sp->sitepolicyrefs[refpos];
				for( GPosition polpos=pr->policy_include_map; polpos; ++polpos)
				{
					{
						/// check for applicable policy 
						GP<PolicyIncludes> pi = pr->policy_include_map[polpos];
						////first this cookie in excludes
						for( GPosition expos=pi->cookie_excludes; expos; ++expos)
						{
							GP<CookieInExclude> cookie = pi->cookie_excludes[expos];
							if( name == cookie->name && patternmatch(cookie->domain, domain) && ( !cookie->value.length() || cookie->value == value ))
								goto bail_out;
						}

						//// check if it is in includes
						for( GPosition inpos=pi->cookie_includes; inpos; ++inpos)
						{
							GP<CookieInExclude> cookie = pi->cookie_includes[inpos];
							if( patternmatch(cookie->name, name) && patternmatch(cookie->domain, domain.upcase()) && ( !cookie->value.length() || cookie->value == value ))
								return pi->policyurl;
						}
					}
					bail_out:;

				}
			}
		}
	}

	return GURL();
}

GURL PolicyReferenceHandler::FindPolicy(const GURL &siteurl, const GString &method)
{
	GCriticalSectionLock lock(&class_lock);

	GURL site;
	if( siteurl.host() == siteurl )
		site = siteurl+"/";
	else
		site = siteurl;

	if( GPosition pos = site_policy_reference_map.contains(site.host()))
	{   //// find the site exists in our map

		GP<SitePolicies> sp = site_policy_reference_map[pos];
		for( GPosition refpos=sp->sitepolicyrefs; refpos; ++refpos)
		{///go through each policy reference to find out we have the required policy url

			GP<PolicyReference> pr  = sp->sitepolicyrefs[refpos];
			for( GPosition polpos=pr->policy_include_map; polpos; ++polpos)
			{/// check for applicable policy 
				
				{
					GP<PolicyIncludes> pi = pr->policy_include_map[polpos];
					//// check for methods
					if( pi->methods.size() )
					{
						for(GPosition mpos = pi->methods; mpos; ++mpos)
						{
							if( pi->methods[mpos] == method )
								break;
						}
						if( !mpos )
							goto bail_out;
					}
					/// first check if this url is in excludes
					for( GPosition expos=pi->excludes; expos; ++expos)
					{  
#ifdef _DEBUG
						const char *temp = (GString)pi->excludes[expos];
#endif
						if(patternmatch(pi->excludes[expos], site) )
							goto bail_out;
					}

					/// then check this url in includes
					for( GPosition inpos=pi->includes; inpos; ++inpos)
					{
#ifdef _DEBUG
						const char *temp = (GString)pi->includes[inpos];
#endif
						if(patternmatch(pi->includes[inpos], site) )
							return pi->policyurl;
					}
				}
			bail_out:;

			}
		}
	}
	
	return GURL();
}

GURL PolicyReferenceHandler::FindPolicyReference(const GURL &site)
{
	GCriticalSectionLock lock(&class_lock);

	return GURL();
}
bool PolicyReferenceHandler::isPolicyReferenceExists(const GURL &policyrefurl)
{
	GCriticalSectionLock lock(&class_lock);

	if( GPosition pos = site_policy_reference_map.contains(policyrefurl.host()))
	{
		GP<SitePolicies> sp = site_policy_reference_map[pos];
		for( GPosition refpos=sp->sitepolicyrefs; refpos; ++refpos)
		{
			GP<PolicyReference> pr  = sp->sitepolicyrefs[refpos];
			if( pr->url == policyrefurl )
				return true;
		}
	}

	return false;
}


bool PolicyReferenceHandler::ParseReferenceFile(const char *buffer, int len, const GURL &policyrefurl)
{
	GCriticalSectionLock lock(&class_lock);

	if( buffer && len)
	{
		TRY
		{
			GP<XmlNode> xmlroot = SaxXmlParse(buffer, len, false /*true*/);//XMLParse(buffer, len);
			return ParseReferenceFile(xmlroot, policyrefurl);
		}
		CATCH(ex)
		{
		}
		ENDCATCH;
	}
	return false;
}
bool PolicyReferenceHandler::ParseReferenceFile(GP<XmlNode> root, const GURL &policyrefurl)
{
	GCriticalSectionLock lock(&class_lock);

	AddDownloadedURL(policyrefurl); /// add to downloaded URLs list
	if( root )
	{
		///should start with "META"
		if( root->name == "META")
		{
			GPosition pos = root->first_child();
			while(pos)
			{
				GP<XmlNode> pcyrefelement = root->next_child(pos);
				ParsePolicyRefrencesElement(pcyrefelement, policyrefurl);
			}
		}
	}
	return false;
}


bool ValidateHintDomain(GString &domain)
{
	TRY
	{
		GURL url = domain;
		GString proto = url.protocol();
		///see if protocol is http or https
		if( (proto == "http") || (proto == "https") )
		{
			GString host = url.host();
			/// see if path is also specified, it is illegal
			if( host.upcase() == domain.upcase() )
			{
				GString dom = url.domain();
				///see if only the first char is * followed by "." or no * allowed else where
				int starpos = dom.search('*');
				if( starpos <= 0 )
				{
					if( starpos < 0 )
						return true;
					if( dom[starpos+1] == '.' && dom.search('*', starpos) <= 0)
						return true;
				}
			}
		}
	}
	CATCH(ex)
	{
	}
	ENDCATCH;
	return false;
}


GP<PolicyReference> PolicyReferenceHandler::ParsePolicyRefrencesElement(GP<XmlNode> node, const GURL &policyrefurl)
{

	if( node && node->name == "POLICY-REFERENCES")
	{
		GP<PolicyReference> pr = new PolicyReference;
		pr->url = policyrefurl;
		GPosition pos = node->first_child();
		while(pos)
		{
			GP<XmlNode> childnode = node->next_child(pos);
			if( childnode->name == "POLICY-REF" )
			{
				GP<PolicyIncludes> pi = ParsePolicyRefElement(childnode, policyrefurl);
				if( pi )
				{
					//pr->policy_include_map[pi->policyurl] = pi;
					pr->policy_include_map.append(pi);
				}
			}
			else if( childnode->name == "EXPIRY" )
			{
				if( GPosition agepos = childnode->attrvalues.contains("MAX-AGE") )
					pr->expiry = atol(childnode->attrvalues[agepos]);
			}
			else if( childnode->name == "HINT" )
			{
				GPosition domainpos;
				if( (domainpos = childnode->attrvalues.contains("DOMAIN")) ||  (domainpos = childnode->attrvalues.contains("SCOPE")))
				{
					if( GPosition pathpos = childnode->attrvalues.contains("PATH") )
					{
						GString domain = childnode->attrvalues[domainpos];
						GString path = childnode->attrvalues[pathpos];
						if( domain.length() && ValidateHintDomain(domain) && path.length() )
							hint_map[domain] = path;
					}
				}
			}
		}
		if( GPosition pos=site_policy_reference_map.contains(policyrefurl.host()))
		{
			GP<SitePolicies> sp = site_policy_reference_map[pos];
			sp->sitepolicyrefs[pr->url] = pr;
		}
		else
		{
			GP<SitePolicies> sp = new SitePolicies;
			sp->sitepolicyrefs[pr->url] = pr;
			site_policy_reference_map[policyrefurl.host()] = sp;
		}

		return pr;
	}
	return NULL;
}

GP<PolicyIncludes> PolicyReferenceHandler::ParsePolicyRefElement(GP<XmlNode> node, const GURL &policyrefurl)
{
	if( node && node->name == "POLICY-REF" )
	{
		if( GPosition pos = node->attrvalues.contains("ABOUT") )
		{
			
			GP<PolicyIncludes> policyincludes = new PolicyIncludes;
			GString strurl = node->attrvalues[pos];
			
			if( strurl[0] == '#' )
				policyincludes->policyurl = GString(policyrefurl)+strurl;
			else if( strurl[0] == '/' )
				policyincludes->policyurl = policyrefurl.host() + strurl;
			else
				policyincludes->policyurl = policyrefurl.base()+strurl;
			
			for(GPosition pos=node->first_child(); pos; )
			{
				GP<XmlNode> child = node->next_child(pos);
				if( child->name == "INCLUDE" )
				{
					///process include
					///check if it is absolute url then invalid
					if( child->content.search("://") == -1 )
					{
						GURL include_url;
						int len = child->content.length();
						if( child->content[len-1] == '/')
							include_url = policyrefurl.host()+child->content+"*";
						else
							include_url = policyrefurl.host()+child->content;

						policyincludes->includes.append(include_url);
					}
				}
				else if( child->name == "EXCLUDE" )
				{
					//process exclude
					///check if it is absolute url then invalid
					if( child->content.search("://") == -1 )
					{
						GURL exclude_url;
						int len = child->content.length();
						if( child->content[len-1] == '/')
							exclude_url = policyrefurl.host()+child->content+"*";
						else
							exclude_url = policyrefurl.host()+child->content;

						policyincludes->excludes.append(exclude_url);
					}
				}
				else if( child->name == "COOKIE-INCLUDE" )
				{
					
					if( GPosition pos = child->attrvalues.contains("NAME") )
					{
						///Process Cookie-Include New syntax
						GP<CookieInExclude> cie = new CookieInExclude();
						cie->name = child->attrvalues[pos];
						if( pos = child->attrvalues.contains("VALUE") )
							cie->value = child->attrvalues[pos];
						if( pos = child->attrvalues.contains("DOMAIN") )
						{
							cie->domain = child->attrvalues[pos].upcase();
							if( cie->domain[0] == '.' )
								cie->domain = cie->domain.substr(1, -1);
						}
						if( pos = child->attrvalues.contains("PATH") )
							cie->path = child->attrvalues[pos];

						policyincludes->cookie_includes.append(cie);
						

					}
					else
					{
						//Process Cookie-Include old syntax
						GString content = child->content;
						if( content.length() )
						{
							GP<CookieInExclude> cie = new CookieInExclude();
							int i = content.search(' ');
							if( i > 0 )
							{
								cie->name = content.substr(0, i);
								int j=i+1;
								i = content.search(' ', j);
								if( i > 0)
								{
									cie->domain = content.substr(j, i-j).upcase();
									if( cie->domain[0] == '.' )
										cie->domain = cie->domain.substr(1, -1);
									cie->path = content.substr(i+1, -1);
								}
							}
							policyincludes->cookie_includes.append(cie);
							
						}
					}
				}
				else if( child->name == "COOKIE-EXCLUDE" )
				{
					if( GPosition pos = child->attrvalues.contains("NAME") )
					{
						///Process Cookie-Exclude New syntax
						GP<CookieInExclude> cie = new CookieInExclude();
						cie->name = child->attrvalues[pos];
						if( pos = child->attrvalues.contains("VALUE") )
							cie->value = child->attrvalues[pos];
						if( pos = child->attrvalues.contains("DOMAIN") )
						{
							cie->domain = child->attrvalues[pos].upcase();
							if( cie->domain[0] == '.' )
								cie->domain = cie->domain.substr(1, -1);
						}
						if( pos = child->attrvalues.contains("PATH") )
							cie->path = child->attrvalues[pos];

						policyincludes->cookie_includes.append(cie);
						

					}
					else
					{
						//Process Cookie-Exclude old syntax
						GString content = child->content;
						if( content.length() )
						{
							GP<CookieInExclude> cie = new CookieInExclude();
							int i = content.search(' ');
							if( i > 0 )
							{
								cie->name = content.substr(0, i);
								int j=i+1;
								i = content.search(' ', j);
								if( i > 0)
								{
									cie->domain = content.substr(j, i-j).upcase();
									if( cie->domain[0] == '.' )
										cie->domain = cie->domain.substr(1, -1);
									cie->path = content.substr(i+1, -1);
								}
							}
							policyincludes->cookie_includes.append(cie);
							
						}
					}
				}
				else if( child->name == "METHOD" )
				{
					policyincludes->methods.append(child->content);
				}

			}
			return policyincludes;

		}
	}
	return NULL;
}


bool PolicyReferenceHandler::isDownloaded(const GURL &url)
{
	GCriticalSectionLock lock(&class_lock);

	if(	downloaded_urls.contains(url) )
		return true;
	else
		return false;

}
bool PolicyReferenceHandler::AddDownloadedURL(const GURL &url)
{
	GCriticalSectionLock lock(&class_lock);

	if( downloaded_urls.contains(url) )
	{
		return false;
	}
	else
	{
		downloaded_urls[url] = true;
		return true;
	}
}
bool PolicyReferenceHandler::RemoveDownloadedURL(const GURL &url)
{
	GCriticalSectionLock lock(&class_lock);

	if( GPosition pos = downloaded_urls.contains(url) )
	{
		downloaded_urls.del(pos);
		return true;
	}
	else
		return false;
}

///////////////////////////////////////////////////////////////////////////
/////////////////General functions                   //////////////////////
///////////////////////////////////////////////////////////////////////////


bool PolicyReferenceHandler::HandlePolicyReferenceURL(const GURL &url, const GURL &siteurl)
{
	if( !isPolicyReferenceExists(url) )
	{
		bool retval = false;

		GString p3pheader;
		char *data = GetURL(url, p3pheader);

		TRY
		{
			if( data )
			{
				ParseReferenceFile(data, strlen(data), url);
				retval = true;
			}
		}
		CATCH(ex)
		{
			
		}
		ENDCATCH;

		if( data )
			delete data;

		GURL p3purl = GetURLFromP3PHeader(p3pheader, url);
		if( !p3purl.is_empty() && url != p3purl)
			HandlePolicyReferenceURL(p3purl, siteurl);


		return retval;
	}
	else
		return true;
}



int PolicyReferenceHandler::HandlePolicyURL(const GURL &realurl, const GURL &siteURL, Result &result)
{
	result.Match = Result::UNKNOWN;
	int retval = 0;
	if( !realurl.is_empty() )
	{
		GURL polurl = realurl;
		result.appliedPolicyURL = polurl;
		GString policyname = SeparateNameAndURL(polurl);
		
		char *policydata=GetURL(polurl);

		///match policy to preferences
		if( policydata )
		{
			TRY
			{
				DArray<char> data;
				if( g_p3p_preferences.ReadPreferenceFile(data) )
				{
					if( P3PEvaluator::P3PEvaluator_Match_Memory(data, data.size(), policydata, strlen(policydata), polurl, policyname, result) == A_OK )
					{
						retval = result.MisMatches.size();
					}
					else
					{
						retval = -1;
					}
				}
				else
					retval = -2;
			}
			CATCH(ex)
			{
				if( policydata )
					delete policydata;
				policydata = 0;
			}
			ENDCATCH;

		}
		else
		{
			retval = -1;
		}

		if( policydata )
			delete policydata;

	}
	return retval;
}



bool PolicyReferenceHandler::GetApplicableHints(const GURL &url, GList<GURL> &hintlist)
{
	GCriticalSectionLock lock(&class_lock);
	bool ret_val = false;

	GURL host = url.host();
	for( GPosition pos = hint_map; pos; ++pos)
	{
		if( patternmatch(hint_map.key(pos), host) )
		{
			hintlist.append(host+hint_map[pos]);
			ret_val = true;
		}
	}
	return ret_val;
}

