#ifndef __XMLNODE_H__
#define __XMLNODE_H__

#include "GContainer.h"
#include "GString.h"
#include "stdio.h"


class XmlNode:public GPEnabled
{

private:
	GP<XmlNode> parent;
	GPList<XmlNode> children;
public:
	GString name;
	GString content;
	GMap<GString, GString> attrvalues;

public:

	XmlNode(GP<XmlNode> par): parent(par){};
	GPosition first_child();
	GP<XmlNode> next_child(GPosition &pos);
	GP<XmlNode> get_parent()
	{
		return parent;
	}
	void add_child(GP<XmlNode> child)
	{
		if( child )
			children.append(child);
	}

	void add_child_copy(GP<XmlNode> child)
	{
		if( child )
			children.append(child->get_copy(this));
	}

	void add_children(GPList<XmlNode> &list);
	void add_children(GP<XmlNode> &node);
	
	void copy_children(GP<XmlNode> &node)
	{
		if( node )
			children = node->children;
	}

	void delete_all_children()
	{
		children.empty();
	}

	void delete_child(GPosition &pos)
	{
		if( pos )
			children.del(pos);
	}

	int delete_named_children(const GString &name);

	bool has_children()
	{
		if( children.size() )
			return true;
	    else
			return false;
	}
	int num_children()
	{
		return children.size();
	}
	GP<XmlNode> get_first_child(GPosition &pos)
	{
		pos = first_child();
		if( pos )
			return next_child(pos);
		else
			return NULL;
	}

	GP<XmlNode> get_next_child(GPosition &pos)
	{
		if( pos )
			return next_child(pos);
		else 
			return NULL;
	}

	GP<XmlNode> get_copy(GP<XmlNode> use_parent);

	void print(FILE *fp, int depth); ///debug routines
	void print(const char *filename);
};


GP<XmlNode> XMLParse(const char *buffer, int len);
GP<XmlNode> XMLParse(const char*filename);
bool InitializeXmlParserFunctions(const GString &path);

#endif