#ifndef __POLICYHANDLER_H__
#define __POLICYHANDLER_H__

//C- Copyright (c) 2001 AT&T Corp.
//C- All rights reserved.
//C- $Id: PolicyHandler.h,v 1.22 2003/01/09 21:50:40 manjula Exp $

#include "GContainer.h"
#include "GString.h"
#include "GURL.h"
#include "XmlNode.h"
#include "P3PEvaluator.h"
#include <time.h>


/* file defines class PolicyHandler maintains policies and policy reference file
   including cache management which is supposed to do the following things
   1. find a policy file exists for a given URL
   2. find a policy exists (URL)
   3. find a policy reference file exists for a given URL
   4. find a policy reference file exists (URL)
   5. parse a given policy reference file and cache its values
   6. find a policy URL for a given URL if it was not already downloaded
   7. 

*/

class CookieInExclude:public GPEnabled
{
public:
	GString name;
	GString value;
	GString domain;
	GString path;
};


class PolicyIncludes:public GPEnabled
{
public:
	GURL policyurl;
	
	GList<GURL> includes;
	GList<GURL> excludes;
	GPList<CookieInExclude> cookie_includes;
	GPList<CookieInExclude> cookie_excludes;
	GList<GString> methods;
};


class CookiePolicy:public GPEnabled
{
public:
	GMap<GP<PolicyIncludes>, bool> policy_include_map;
};


class PolicyReference:public GPEnabled
{
public:
	

	GURL url; ///policy ref URL
	long expiry;
	GPList<PolicyIncludes> policy_include_map;
};

class SitePolicies:public GPEnabled
{
public:
	////map for policy reference URL and policy reference structure
	GMap<GURL, GP<PolicyReference> > sitepolicyrefs;
};

class CacheData:public GPEnabled
{
public:
	time_t expiry;
	GString filename;
	CacheData(const GString &_filename, const time_t _expiry):filename(_filename), expiry(_expiry) {};
	CacheData():expiry(0){};
};

class PolicyReferenceHandler
{
	////map for site host name and SitePolicies structure
	static GMap<GURL, GP<SitePolicies> > site_policy_reference_map;

	//// map for policy url and policy data
	static GMap<GURL, GP<CacheData> > policy_cache;
	//// hints collected
	static GMap<GString, GString> hint_map; //// its is domain to path map
	static GMap<GURL, bool> downloaded_urls;

	static GCriticalSection class_lock;

public:
	GURL FindPolicy(const GURL &site, const GString &method);
	GURL FindPolicyForCookie(const GString &name, const GString &value, GString domain, GString path);
	GURL FindPolicyReference(const GURL &site);
	bool isPolicyReferenceExists(const GURL &policyrefurl);
	////cache functions
private:
	char *GetURLFromCache(const GURL &policyurl, bool deleteifexpired=true);
	bool CacheURL(const GURL &policyurl, const char *policy, time_t expiry=0);
	
public:
	bool isURLExistsInCache(const GURL &policyurl);
	GString GetLocalPathFromCache(const GURL &policyurl);
	char *GetURL(const GURL &url, GString &header, bool download=true, bool deleteifexpired=true);
	char *GetURL(const GURL &url, bool download=true, bool deleteifexpired=true);

	////////////////////////////////////////////////
	bool ParseReferenceFile(const char *buffer, int len, const GURL &policyrefurl);
	bool ParseReferenceFile(GP<XmlNode> root, const GURL &policyrefurl);

	////////////////////////////////////////////////
	bool isDownloaded(const GURL &url);
	bool AddDownloadedURL(const GURL &url);
	bool RemoveDownloadedURL(const GURL &url);

private:
	GP<PolicyReference> ParsePolicyRefrencesElement(GP<XmlNode> node, const GURL &policyrefurl);
	GP<PolicyIncludes> ParsePolicyRefElement(GP<XmlNode> node, const GURL &policyrefurl);


public:
////////////////////////////////////////////////////////////////
/////////////////////General Functions
////////////////////////////////////////////////////////////////

	bool HandlePolicyReferenceURL(const GURL &url, const GURL &siteurl);
	int HandlePolicyURL(const GURL &polurl, const GURL &siteURL,
			    DArray<char> &data, Result &result);
	bool GetApplicableHints(const GURL &url, GList<GURL> &hintlist);

	static void SavePolicyCache();
	static void LoadPolicyCache();

};

///global policy and policy reference handler
extern PolicyReferenceHandler g_policy_handler;





#endif
