//C- $Id: XmlFile.cpp,v 1.3 2003/01/30 13:50:43 praveen Exp $

#include "XmlFile.h"
#include <stdio.h>



XmlOutFile::XmlOutFile(GString path):depth(0)
{
	if( path.length() )
	{
		filename = path;
		fp = fopen(filename, "w");
		if( !fp )
			THROW("Could not open file for writing, File could be locked by other program");

		fprintf(fp, "<?xml version=\"1.0\"?>\n");
	}
	else
		THROW("Empty file name");
}

XmlOutFile::~XmlOutFile()
{
	if( fp )
	{
		fclose(fp);
	}
}

void XmlOutNode::println(const char *string)
{
	if( this->name.length() )
	{
		print(string);
		fprintf(file.fp, "\n");
	}
	else
		THROW("Node name not set, use setNodeName() first");
}

void XmlOutNode::print(const char *string)
{
	if( this->name.length() )
	{
		if( add_depth )
		{
			for(int i=0; i<file.depth; i++)
				fprintf(file.fp, "  ");
		}
		fprintf(file.fp, string);
	}
	else
		THROW("Node name not set, use setNodeName() first");
}


XmlOutNode::~XmlOutNode()
{
	if(!empty && name.length())
	{
		if( add_depth )
		{
			for(int i=0; i<file.depth; i++)
				fprintf(file.fp, "  ");
		}

		fprintf(file.fp, "</%s>", (const char *)name);
		if( add_newline )
			fprintf(file.fp, "\n");

	}
	file.depth--;
	if( file.depth < 0 )
		THROW("Depth is negative");
}

void XmlOutNode::setNodeName(GString name, GList<GString> &att, GList<GString> &val)
{
	if( this->name.length() )
		THROW("Node name already set");

	if( name.length() )
	{
		this->name = name;
		if( add_depth )
		{
			for(int i=0; i<file.depth; i++)
				fprintf(file.fp, "  ");
		}

		fprintf(file.fp, "<%s ", (const char *)name);

		GPosition pos1, pos2;
		
		pos2 = val;

		for(pos1 = att; pos1; ++pos1)
		{
			fprintf(file.fp, "%s", (const char*)att[pos1]);
			if( pos2 )
			{
				fprintf(file.fp, "=\"%s\"", (const char*)val[pos2]);
				++pos2;
			}
			fprintf(file.fp, " ");
		}

		if( empty )
		{
			
			fprintf(file.fp, "/>");
			if( add_newline )
				fprintf(file.fp, "\n");
		}
		else
		{
			
			fprintf(file.fp, ">");
			if( add_newline )
				fprintf(file.fp, "\n");
		}
	}
}


void XmlOutNode::setNodeName(GString name, GMap<GString, GString> &attrvalues)
{
	GList<GString> att;
	GList<GString> vals;

	for(GPosition pos=attrvalues; pos; ++pos)
	{
		att.append(attrvalues.key(pos));
		vals.append(attrvalues[pos]);
	}

	setNodeName(name, att, vals);
}

void XmlOutNode::setNodeName(GString name)
{
	if( name.length() )
	{
		this->name = name;
		if( add_depth )
		{
			for(int i=0; i<file.depth; i++)
				fprintf(file.fp, "  ");
		}

		fprintf(file.fp, "<%s ", (const char *)name);
		if( empty )
		{
			
			fprintf(file.fp, "/>");
			if( add_newline )
				fprintf(file.fp, "\n");
		}
		else
		{
			
			fprintf(file.fp, ">");
			if( add_newline )
				fprintf(file.fp, "\n");
		}
	}
}


void XmlOutNode::add_comment(const char * comment)
{
	if( strlen(comment) )
	{
		if( add_depth )
		{
			for(int i=0; i<file.depth; i++)
				fprintf(file.fp, "  ");
		}
		fprintf(file.fp, "<!--%s-->", comment);
		if( add_newline )
				fprintf(file.fp, "\n");

	}
}




