#include "XmlNode.h"
#include "expat.h"
#include <stdio.h>
#include <iostream.h>

static bool CheckXmlParserModuleAvailable() {
    return true;
}

void XmlNode::print(FILE *fp, int depth) {
  GPosition pos;

  depth += 2;
  if(!fp)
    return;

  for(int i=0; i<depth; i++)
    fprintf(fp, " ");

  /// print name
  fprintf(fp, "<%s", (const char*) name);
  for(pos=attrvalues; pos; ++pos) {
    fprintf(fp, "  %s=\"%s\"", (const char*)attrvalues.key(pos), (const char*)attrvalues[pos]);
  }
	
	

  if(content.length() > 0 || children.size() > 0) {
    fprintf(fp, ">");
    fprintf(fp, (const char*)content);
    for(pos = children; pos; ++pos)
      children[pos]->print(fp, depth);
  }

  if(content.length() > 0 || children.size() > 0) {
    for(int i=0; i<depth; i++)
      fprintf(fp, " ");
    fprintf(fp, "</%s>", (const char*) name);
  }
  else
    fprintf(fp, "/>");
  fprintf(fp, "\n");

  depth -= 2;
	


	
}


void XmlNode::print(const char *filename) {
  FILE *fp;
  if(fp = fopen(filename, "w")) {
    print(fp, 0);
    fclose(fp);
  }
}

int XmlNode::delete_named_children(const GString &name) {
  int num=0;
  for(GPosition pos=children; pos;) {
    GPosition tmppos = pos;
    ++pos;
    if(children[tmppos]->name == name) {
      children.del(tmppos);
      num++;
    }
  }
  return num;
}
			

void XmlNode::add_children(GPList<XmlNode> &list) {
  for(GPosition pos=list; pos; ++pos)
    children.append(list[pos]);
}

void XmlNode::add_children(GP<XmlNode> &node) {
  for(GPosition pos=node->children; pos; ++pos)
    children.append(node->children[pos]);
}

GP<XmlNode> XmlNode::get_copy(GP<XmlNode> use_parent) {
  GP<XmlNode> node = new XmlNode(use_parent);
  node->name = name;
  node->content = content;
  node->attrvalues = attrvalues;
  GPosition pos;
  GP<XmlNode> child = get_first_child(pos);
  while(child) {
    node->children.append(child->get_copy(node));

    child = get_next_child(pos);
  }

  return node;
}

GPosition XmlNode::first_child() {
  GPosition pos = children;
  return pos;
}

GP<XmlNode> XmlNode::next_child(GPosition &pos) {
  if(pos) {
    GP<XmlNode> child = children[pos];
    ++pos;
    return child;
  }
  else
    return GP<XmlNode>();
}

typedef struct _XMLParseData {
  XmlNode *curnode;
  GP<XmlNode> root;
} XMLParseData;



static void XML_Start(void *userdata, const XML_Char * name,
                      const XML_Char ** atts) {
  XMLParseData *parsedata = (XMLParseData*)userdata;
  GP<XmlNode> node = new XmlNode(parsedata->curnode);

  if(!parsedata->curnode) {
    parsedata->curnode = node;
    parsedata->root = node;
  }
  else {
    ///add node as child
    GP<XmlNode> parent = parsedata->curnode;
    parent->add_child(node);
  }
	
  node->name = GString(name).upcase();
  for(int i=0; atts[i]; i+=2) {
    node->attrvalues[GString(atts[i]).upcase()] = atts[i+1];
    if(!atts[i+1])
      break;
  }

  parsedata->curnode = node;
}

static void XML_CharData(void *userdata, const XML_Char *s,int len) {
  XMLParseData *parsedata = (XMLParseData*)userdata;
  XmlNode *node;
  node = parsedata->curnode;
  if(node) {
    GString str(s, len);
    node->content += str;
  }

}

static void XML_End(void *userdata, const XML_Char * name) {

  XMLParseData *parsedata = (XMLParseData*)userdata;
  XmlNode* node = parsedata->curnode;
  if(node) {
    node = node->get_parent();
    parsedata->curnode = node;
  }

}

//parses XML data and returns root node of the parsed tree
GP<XmlNode> XMLParse(const char *buffer, int len) {
  ///lockit so that only one thread can enter it at a time

  if(!CheckXmlParserModuleAvailable())
    THROW("Error: Xml Parser Module not loaded");

  XML_Parser parser;

  parser = XML_ParserCreate(NULL/*"UTF-8"*/);
  XML_SetElementHandler(parser, XML_Start, XML_End);
  XML_SetCharacterDataHandler(parser, XML_CharData);

  XMLParseData parsedata;
  parsedata.curnode = parsedata.root = NULL;
  XML_SetUserData(parser, (void*)&parsedata);
	
  if (XML_Parse(parser, buffer, len, 1) == 0) {
    XML_ParserFree(parser);
    THROW("XML parse error");
  }

  XML_ParserFree(parser);
  return parsedata.root;

}

GP<XmlNode> XMLParse(const char*filename) {
  if(filename) {
    FILE *fp;
    if(fp = fopen(filename, "rb")) {
      fseek(fp, 0, SEEK_END);
      int length = ftell(fp);
      fseek(fp, 0, SEEK_SET);
      GString strbuffer;
      char *buffer = strbuffer.getbuf(length+1);
      if(buffer) {
	fread(buffer, length, 1, fp);
	fclose(fp);
	return XMLParse(buffer, length);
      }
      fclose(fp);
    }
  }
  return NULL;
}

